#include "spadfs.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#define OLD_SYMLINKS
#define ERR_PTR
#endif

#ifndef OLD_SYMLINKS
static void *spadfs_follow_link(struct dentry *dentry, struct nameidata *nd)
#else
static int spadfs_follow_link(struct dentry *dentry, struct nameidata *nd)
#endif
{
	SPADFNODE *f = spadfnode(dentry->d_inode);
	struct fnode_ea *lnk = GET_EA((struct fnode_ea *)f->ea, f->ea_size,
				      EA_SYMLINK_MAGIC, FNODE_EA_MAGIC_MASK);
	unsigned len;
	char *str;

	if (unlikely(!lnk)) {
		spadfs_error(f->fs, TXFLAGS_EA_ERROR,
			"can't find symlink extended attribute on fnode %Lx/%x",
			(unsigned long long)f->fnode_block,
			f->fnode_pos);
		return ERR_PTR(-EFSERROR);
	}

	if (unlikely(lnk == GET_EA_ERROR)) {
		spadfs_error(f->fs, TXFLAGS_FS_ERROR,
			"error parsing extended attributes on symlink fnode %Lx/%x",
			(unsigned long long)f->fnode_block,
			f->fnode_pos);
		return ERR_PTR(-EFSERROR);
	}

	len = SPAD2CPU32_LV(&lnk->magic) >> FNODE_EA_SIZE_SHIFT;
	str = kmalloc(len + 1, GFP_NOFS);	/* ??? GFP_KERNEL */
	if (unlikely(!str))
		return ERR_PTR(-ENOMEM);

	memcpy(str, lnk + 1, len);
	str[len] = 0;
	nd_set_link(nd, str);
#ifndef OLD_SYMLINKS
	return str;
#else
	return 0;
#endif
}

#ifndef OLD_SYMLINKS
static void spadfs_put_link(struct dentry *dentry, struct nameidata *nd, void *str)
{
	kfree(str);
}
#else
static void spadfs_put_link(struct dentry *dentry, struct nameidata *nd)
{
	char *str = nd_get_link(nd);
	kfree(str);
}
#endif

static int spadfs_symlink_setattr(struct dentry *dentry, struct iattr *iattr)
{
	iattr->ia_valid &= ~(ATTR_MODE | ATTR_UID | ATTR_GID | ATTR_SIZE);
	return spadfs_file_setattr(dentry, iattr);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
const
#endif
struct inode_operations spadfs_symlink_iops = {
	.readlink = generic_readlink,
	.follow_link = spadfs_follow_link,
	.put_link = spadfs_put_link,
	.setattr = spadfs_symlink_setattr,
	.getattr = spadfs_getattr,
#ifdef SPADFS_XATTR
	.setxattr = generic_setxattr,
	.getxattr = generic_getxattr,
	.listxattr = generic_listxattr,
	.removexattr = generic_removexattr,
#endif
};
